// $Id: Plugins.cs 25 2008-04-20 14:51:40Z arcao $
//
// Copyright (C) 2008 Martin "Arcao" Sloup <arcao@arcao.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Reflection;
using System.Xml;
using OSD.PluginBase;

namespace OSD.Program
{
    /// <summary>
    /// Trida udrzujici vsechny informace o DownloadPluginech
    /// </summary>
    public class Plugins : IEnumerable<DownloadPlugin>
    {
        /// <summary>
        /// adresar, kde se nachazeji pluginy programu
        /// </summary>
        public const string PLUGIN_PATH = @"plugins";

        private static Plugins plugins;
        private readonly List<DownloadPlugin> pluginList;
        private readonly Settings settings;

        private Plugins(Settings settings)
        {
            pluginList = new List<DownloadPlugin>();
            this.settings = settings;
        }

        /// <summary>
        /// Indexer, ktery po zadani indexu DownloadPluginu vraci aktivni DownloadPlugin
        /// </summary>
        /// <param name="index">index DownloadPluginu</param>
        /// <returns>DownloadPlugin na prislusnem indexu</returns>
        public DownloadPlugin this[int index]
        {
            get { return pluginList[index]; }
        }

        /// <summary>
        /// Vraci pocet aktivnich DownloadPluginu
        /// </summary>
        public int Count
        {
            get { return pluginList.Count; }
        }

        #region Enumerators

        IEnumerator<DownloadPlugin> IEnumerable<DownloadPlugin>.GetEnumerator()
        {
            return pluginList.GetEnumerator();
        }

        /// <summary>
        /// Vraci enumerator aktvnich DownloadPluginu
        /// </summary>
        /// <returns>enumerator DownloadPluginu</returns>
        public IEnumerator GetEnumerator()
        {
            return ((IEnumerable<DownloadPlugin>) this).GetEnumerator();
        }

        #endregion

        /// <summary>
        /// Handler udalosti, ktera se vykona, po zavolani metody Save.
        /// </summary>
        public event EventHandler ReloadPlugins;

        /// <summary>
        /// Vraci statickou instanci teto tridy a zaroven pokud jeste neexistuje zadna takova instance teto tridy, provede nacteni DownloadPluginu, dle nastaveni aplikace.
        /// </summary>
        /// <returns>statickou instanci teto tridy</returns>
        public static Plugins GetInstance()
        {
            if (plugins == null)
                plugins = new Plugins(Settings.getInstance());
            return plugins;
        }

        /// <summary>
        /// Vraci true, pokud existuje DownloadPlugin pro zadane url.
        /// </summary>
        /// <param name="url">url, ktere se ma testovat</param>
        /// <returns>true, pokud existuje DownloadPlugin pro zadane url</returns>
        public bool isPluginFor(string url)
        {
            foreach (DownloadPlugin item in pluginList)
                if (item.IsSuitableUrl(url)) return true;
            return false;
        }

        /// <summary>
        /// Vraci instanci Download tridy, ktera je schopna stahnout zadane url. Pokud zadna takova neni, vraci null.
        /// </summary>
        /// <param name="url">url, ke ktere shanime instanci Download tridy</param>
        /// <returns>instance Download tridy, pokud takova existuje, jinak null</returns>
        public Download GetDownloadFor(string url)
        {
            foreach (DownloadPlugin item in pluginList)
                if (item.IsSuitableUrl(url)) return item.GetDownloadForUrl(url);
            return null;
        }

        /// <summary>
        /// Vraci instanci DownloadPluginu, ktery se nachazi v zadanem souboru. Nazev souboru musi mit spravny format.
        /// Priklad: 
        ///   soubor: cesta/MujServer.DownloadPlugin.dll
        ///   trida: OSD.Plugins.MujServer.DownloadPlugin
        /// </summary>
        /// <param name="filename">nazev knihovny obsahujici DownloadPlugin</param>
        /// <returns>instance DownloadPluginu</returns>
        public static DownloadPlugin GetPlugin(string filename)
        {
            try
            {
                string className;

                if (!File.Exists(filename)) return null;

                if (filename.LastIndexOf("\\") > -1)
                    className = filename.Substring(filename.LastIndexOf("\\") + 1);
                else if (filename.LastIndexOf("/") > -1)
                    className = filename.Substring(filename.LastIndexOf("/") + 1);
                else
                    className = filename;

                if (className.EndsWith(".dll"))
                    className = className.Substring(0, className.Length - 4);
                if (!className.EndsWith(".DownloadPlugin")) return null;

                Type objType = null;
                // load it
                Assembly ass;
                ass = Assembly.LoadFrom(filename);
                if (ass != null)
                    objType = ass.GetType("OSD.Plugins." + className);
                if (objType != null)
                    return (DownloadPlugin) Activator.CreateInstance(objType, Settings.getInstance());
            }
            catch {}
            return null;
        }

        /// <summary>
        /// Vraci seznam instanci DownloadPluginu nachazejici se v adresari PLUGIN_PATH
        /// </summary>
        /// <returns>seznam instanci DownloadPluginu</returns>
        public static List<DownloadPlugin> GetAvailablePlugins()
        {
            string[] pluginFiles = Directory.GetFiles(PLUGIN_PATH, "*.dll");

            List<DownloadPlugin> ret = new List<DownloadPlugin>();

            foreach (string file in pluginFiles)
            {
                DownloadPlugin dp = GetPlugin(file);

                if (dp != null)
                    ret.Add(dp);
            }

            return ret;
        }

        #region Load / Save methods

        /// <summary>
        /// Z nastaveni nacte seznam aktivnich DownloadPluginu
        /// </summary>
        public void Load()
        {
            XmlNodeList nl = settings.getNodes("//osd/plugins/plugin");
            foreach (XmlNode n in nl)
            {
                if (Boolean.Parse(n.Attributes["active"].Value))
                {
                    String pluginName = n.SelectSingleNode("plugin_name").InnerText;
                    DownloadPlugin dp = GetPlugin(string.Format(@"{0}\{1}.dll", PLUGIN_PATH, pluginName));
                    if (dp == null)
                        continue;

                    pluginList.Add(dp);
                }
            }
        }

        /// <summary>
        /// Do nastaveni programu ulozi seznam aktivnich DownloadPluginu
        /// </summary>
        public void Save()
        {
            XmlNodeList nl;
            XmlNode pluginNode;

            //skontrolujeme, zda parent node existuje, pokud ne, zalozimeho
            if (settings.getNode("osd/plugins") == null)
                settings.createSelectNodes(settings.XmlDocument, "osd/plugins");

            //projedeme vsechny dostupne pluginy a pridame je do nastaveni, pokud tam jiz nejsou
            List<DownloadPlugin> allPlugins = GetAvailablePlugins();
            foreach (DownloadPlugin plugin in allPlugins)
            {
                string pluginPath = settings.getPluginNodePath(plugin.PluginClassName);
                if (pluginPath.Length == 0)
                {
                    pluginNode = settings.XmlDocument.CreateElement("plugin");
                    pluginNode.AppendChild(settings.XmlDocument.CreateElement("plugin_name")).AppendChild(
                        settings.XmlDocument.CreateTextNode(plugin.PluginClassName));

                    //pridame node
                    settings.getNode("osd/plugins").AppendChild(pluginNode);
                }
                
            }

            //nejprve deaktivujeme vsechny pluginy
            nl = settings.getNodes("osd/plugins/plugin");
            foreach (XmlNode n in nl)
                ((XmlElement) n).SetAttribute("active", Boolean.FalseString);

            // nyni nastavime aktivni pluginy jen ty, ktere chceme
            foreach (DownloadPlugin plugin in pluginList)
            {
                String pluginName = plugin.PluginClassName;

                //najdeme node, pokud existuje, jinak ho zalozime
                pluginNode = null;

                nl = settings.getNodes("osd/plugins/plugin");
                foreach (XmlNode n in nl)
                {
                    if (n.SelectSingleNode("plugin_name").InnerText == pluginName)
                    {
                        pluginNode = n;
                        break;
                    }
                }
                if (pluginNode == null)
                {
                    pluginNode = settings.XmlDocument.CreateElement("plugin");
                    pluginNode.AppendChild(settings.XmlDocument.CreateElement("plugin_name")).AppendChild(
                        settings.XmlDocument.CreateTextNode(pluginName));

                    //pridame node
                    settings.getNode("osd/plugins").AppendChild(pluginNode);
                }

                //nastavime plugin jako aktivni
                ((XmlElement) pluginNode).SetAttribute("active", Boolean.TrueString);
            }

            //posleme event, ze jsme reloadnuly pluginy
            if (ReloadPlugins != null)
                ReloadPlugins(this, null);
        }

        #endregion

        #region Add / Remove / Set / Clear methods

        /// <summary>
        /// Mezi aktivni DownloadPluginy prida DownloadPlugin nachazejici se v knihovne zadane jako nazev souboru. Zadany soubor musi mit spravny nazev. Viz metoda <see cref="GetPlugin">GetPlugin</see>
        /// </summary>
        /// <param name="filename">knihovna, kde se nachazi DownloadPlugin</param>
        /// <returns>true, pokud se DownloadPlugin podarilo pridat</returns>
        public bool AddPlugin(string filename)
        {
            DownloadPlugin dp = GetPlugin(filename);
            if (dp != null && !pluginList.Contains(dp))
            {
                pluginList.Add(dp);
                return true;
            }
            return false;
        }

        /// <summary>
        /// Mezi aktivni DownloadPluginy prida DownloadPlugin zadany v parametru
        /// </summary>
        /// <param name="dp">DownloadPlugin, ktery se ma pridat</param>
        /// <returns>true, pokud se DownloadPlugin podarilo pridat</returns>
        public bool AddPlugin(DownloadPlugin dp)
        {
            if (pluginList.Contains(dp)) return false;
            pluginList.Add(dp);
            return true;
        }

        /// <summary>
        /// Mezi aktivni DownloadPluginy prida DownloadPlugin zadany jako cast nazvu tridy.
        /// 
        /// Napriklad pro DownloadPlugin OSD.Plugins.MujServer.DownloadPlugin je spravny parametr:
        ///   MujServer.DownloadPlugin
        /// 
        /// </summary>
        /// <param name="className">nazev tridy</param>
        /// <returns>true, pokud se DownloadPlugin podarilo pridat</returns>
        public bool AddPluginByClassName(string className)
        {
            return AddPlugin(PLUGIN_PATH + className + ".dll");
        }

        /// <summary>
        /// Z aktivnich DownloadPluginu odebere DownloadPlugin zadany jako cast nazvu tridy.
        /// 
        /// Napriklad pro DownloadPlugin OSD.Plugins.MujServer.DownloadPlugin je spravny parametr:
        ///   MujServer.DownloadPlugin
        /// 
        /// </summary>
        /// <param name="className">nazev tridy</param>
        public void RemovePluginByClassName(string className)
        {
            int i = 0;
            while (i < pluginList.Count)
            {
                if (pluginList[i].PluginClassName == className)
                {
                    pluginList.RemoveAt(i);
                    break;
                }
                i++;
            }
        }

        /// <summary>
        /// Mezi aktivni DownloadPluginy prida DownloadPluginy zadane prostrednictvim enumeratoru.
        /// </summary>
        /// <param name="collection">enumerator s DownloadPluginy</param>
        public void AddRange(IEnumerable<DownloadPlugin> collection)
        {
            pluginList.AddRange(collection);
        }

        /// <summary>
        /// Jako aktivni DownloadPluginy nastavi DownloadPluginy zadane prostrednictvim enumeratoru.
        /// </summary>
        /// <param name="collection">enumerator s DownloadPluginy</param>
        public void SetPlugins(IEnumerable<DownloadPlugin> collection)
        {
            pluginList.Clear();
            if (collection != null) pluginList.AddRange(collection);
        }

        /// <summary>
        /// Smaze vsechny aktivni DownloadPluginy
        /// </summary>
        public void Clear()
        {
            pluginList.Clear();
        }

        #endregion
    }
}