﻿// $Id: Download.cs 25 2008-04-20 14:51:40Z arcao $
//
// Copyright (C) 2008 Martin "Arcao" Sloup <arcao@arcao.com>
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Threading;

namespace OSD.PluginBase
{
    /// <summary>
    /// Vycet moznych stavu, ktere mohou nastat pri stahovani
    /// </summary>
    public enum DownloadStatus
    {
        /// <summary>
        /// Pripojuje se k serveru
        /// </summary>
        Connecting = 1,
        /// <summary>
        /// Probiha stahovani
        /// </summary>
        Downloading = 2,
        /// <summary>
        /// Probiha cekani na zacatek stahovani
        /// </summary>
        Waiting = 3,
        /// <summary>
        /// Vyprselo spojeni se serverem. Take oznacuje stav, kdy byl prekrocen denni limit (pak je nastaven WaitingTime na cas, po kterym se bude zase moci stahovat)
        /// </summary>
        Timeout = 4,
        /// <summary>
        /// Nastala chyba po vyplneni captcha obrazku
        /// </summary>
        CaptchaFailed = 5,
        /// <summary>
        /// Nastala chyba pri ukladani, ci pri stahovani souboru
        /// </summary>
        IOError = 6,
        /// <summary>
        /// Soubor byl na serveru smazan
        /// </summary>
        Deleted = 7,
        /// <summary>
        /// Soubor byl uspesne stahnut
        /// </summary>
        Downloaded = 8
    }

    ///<summary>
    /// Zakladni trida nutna k implementaci stahovaci tridy. Budouci trida musi doimplementovat metodu DownloadThread() kde se nachazi cela logika stahovani
    ///</summary>
    public abstract class Download : IDisposable
    {
        private readonly DownloadPlugin parrent;
        private readonly Thread thread;
        private readonly string url;

        /// <summary>
        /// Objekt WebDownloaderu. Pri kazdem znovupouziti musi byt zavolano wd.Clear(), aby se smazaly nektere hlavicky predchoziho pozadavku.
        /// </summary>
        protected WebDownloader wd = new WebDownloader();

        ///<summary>
        /// Konstruktor stahovaci tridy
        ///</summary>
        ///<param name="url">url souboru, ktery se bude stahovat</param>
        ///<param name="parrent">trida DownloadPluginu, kteremu nalezi tato trida</param>
        public Download(string url, DownloadPlugin parrent)
        {
            this.parrent = parrent;

            if (!url.StartsWith("http://", StringComparison.InvariantCultureIgnoreCase))
                url = "http://" + url;

            this.url = url;

            LoadSettings();

            thread = new Thread(this.DownloadThread);
            thread.IsBackground = true;
            thread.Start();
        }

        ///<summary>
        /// Vraci stav stahovani
        ///</summary>
        public abstract DownloadStatus Status { get; }

        ///<summary>
        /// Vraci stav stahovani v textove podobe. Muze take obsahovat dalsi texty, ktery vratil server, odkud se stahuje. 
        ///</summary>
        public abstract String StatusText { get; }

        ///<summary>
        /// Vraci velikost stahovaneho souboru. Pokud jeste samotny proces stahovani nezacal vraci 0.
        ///</summary>
        public abstract long Size { get; }

        ///<summary>
        /// Vraci aktualni stav stahovani v procentech.
        ///</summary>
        public int Percents
        {
            get { return (Size == 0) ? 0 : (int) (((float) Position/Size)*100); }
        }

        ///<summary>
        /// Vraci pocet bajtu, ktere jiz byli stahnuty
        ///</summary>
        public abstract long Position { get; }

        ///<summary>
        /// Vraci doporuceny cas, ktery je nutno pockat pro dalsi akci. Nastaveni tohoto gettru se provede pri Status == DownloadStatus.Waiting a Status == DownloadStatus.Timeout.
        ///</summary>
        public abstract int WaitTime { get; }

        ///<summary>
        /// Vraci url souboru, ktery se bude stahovat
        ///</summary>
        public string Url
        {
            get { return url; }
        }

        ///<summary>
        /// Vraci vlastnika (DownloadPlugin) teto tridy
        ///</summary>
        public DownloadPlugin Parrent
        {
            get { return parrent; }
        }

        #region IDisposable Members

        ///<summary>
        ///Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        ///</summary>
        ///<filterpriority>2</filterpriority>
        public void Dispose()
        {
            Stop();
            ClearResources();

            if (wd == null) return;
            wd.Dispose();
            wd = null;
        }

        #endregion

        ///<summary>
        /// Funkce starajici se smazani pouzitych zdroju. Zavola se automaticky po zruseni tridy.
        ///</summary>
        protected abstract void ClearResources();

        ///<summary>
        /// Hlavni metoda starijici se o stahovani. Metoda je spustena v samostatnem vlakne. Zdedena trida musi tuto metodu implementovat.
        ///</summary>
        protected abstract void DownloadThread();

        /// <summary>
        /// Nacte vychozi nastaveni objektu WebDownloaderu (wd).
        /// </summary>
        protected void LoadSettings()
        {
            int readBufferLength = parrent.Settings.getInt("osd/settings/read_buffer_length", 64)*1024;
            wd.ReadBufferLength = readBufferLength > 0 ? readBufferLength : WebDownloader.READ_BUFFER_LENGTH;
            wd.UserAgent = "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)";
            wd.Accept =
                "text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5";
            wd.Headers.Add("Accept-Language", "cs,en-us;q=0.7,en;q=0.3");
            wd.Headers.Add("Accept-Charset", "windows-1250,utf-8;q=0.7,*;q=0.7");
        }

        ///<summary>
        /// Ukonci proces stahovani. Tim padem ukonci Metodu DownloadThread().
        ///</summary>
        public void Stop()
        {
            try
            {
                thread.Abort();
                thread.Join();
            }
            catch {}
        }

        #region Events

        ///<summary>
        /// Handler pro udalost, ktera nastane pri zmene stavu behem stahovani
        ///</summary>
        public event EventHandler StatusChanged;

        /// <summary>
        /// Handler pro udalost, ktera nastane pri zmene poctu stahnutych dat
        /// </summary>
        public event EventHandler PositionChanged;

        /// <summary>
        /// Handler pro udalost, ktera nastane pri dokonceni stahovani
        /// </summary>
        public event EventHandler DownloadComplete;

        /// <summary>
        /// Handler pro udalost, ktera nastane, pokud se stahovani nepovede
        /// </summary>
        public event EventHandler DownloadFailed;

        /// <summary>
        /// Zavola handler udalosti StatusChanged
        /// </summary>
        /// <param name="e">objekt s argumenty, ktere se prenesou v udalosti</param>
        protected virtual void OnStatusChanged(DownloadStatusChangedEventArgs e)
        {
            if (StatusChanged != null) StatusChanged(this, e);
        }

        /// <summary>
        /// Zavola handler udalosti PositionChanged
        /// </summary>
        /// <param name="e">objekt s argumenty, ktere se prenesou v udalosti</param>
        protected virtual void OnPositionChanged(PositionChangedEventArgs e)
        {
            if (PositionChanged != null) PositionChanged(this, e);
        }

        /// <summary>
        /// Zavola handler udalosti DownloadComplete. Udalost je spustena na novem vlakne.
        /// </summary>
        /// <param name="e">objekt s argumenty, ktere se prenesou v udalosti</param>
        protected virtual void OnDownloadComplete(DownloadCompleteEventArgs e)
        {
            if (DownloadComplete != null) DownloadComplete.BeginInvoke(this, e, null, null);
        }

        /// <summary>
        /// Zavola handler udalosti DownloadFailed. Udalost je spustena na novem vlakne.
        /// </summary>
        /// <param name="e">objekt s argumenty, ktere se prenesou v udalosti</param>
        protected virtual void OnDownloadFailed(DownloadFailedEventArgs e)
        {
            if (DownloadFailed != null) DownloadFailed.BeginInvoke(this, e, null, null);
        }

        #endregion
    }

    /// <summary>
    /// Trida zastresujici hodnotu argumentu u udalosti DownloadComplete
    /// </summary>
    public class DownloadCompleteEventArgs : EventArgs
    {
        private readonly string filename;
        private readonly string url;

        /// <summary>
        /// Konstruktor tridy
        /// </summary>
        /// <param name="url">url souboru, ktery byl stazen</param>
        /// <param name="filename">cesta vcetne nazvu souboru, kam byl souubor stazen</param>
        public DownloadCompleteEventArgs(string url, string filename)
        {
            this.url = url;
            this.filename = filename;
        }

        /// <summary>
        /// Vraci url souboru, ktery byl stazen
        /// </summary>
        public string Url
        {
            get { return url; }
        }

        /// <summary>
        /// Vraci cestu vcetne nazvu souboru, kam byl souubor stazen
        /// </summary>
        public string Filename
        {
            get { return filename; }
        }
    }

    /// <summary>
    /// Trida zastresujici hodnotu argumentu u udalosti DownloadStatusChanged
    /// </summary>
    public class DownloadStatusChangedEventArgs : EventArgs
    {
        private readonly DownloadStatus status;
        private readonly string statusText;

        /// <summary>
        /// Konstruktor tridy
        /// </summary>
        /// <param name="status">stav, ktery nastal pri stahovani</param>
        /// <param name="statusText">textova verze stavu</param>
        public DownloadStatusChangedEventArgs(DownloadStatus status, string statusText)
        {
            this.status = status;
            this.statusText = statusText;
        }

        /// <summary>
        /// Vraci stav, ktery nastal pri stahovani
        /// </summary>
        public DownloadStatus Status
        {
            get { return status; }
        }

        /// <summary>
        /// Vraci textovou verzi stavu
        /// </summary>
        public string StatusText
        {
            get { return statusText; }
        }
    }

    /// <summary>
    /// Trida zastresujici hodnotu argumentu u udalosti DownloadStatusChange s urcenim casu pro timeout
    /// </summary>
    public class TimeoutStatusChangeEventArgs : DownloadStatusChangedEventArgs
    {
        private readonly int suggestedWaiting = 0;

        /// <summary>
        /// Konstruktor tridy
        /// </summary>
        /// <param name="status">stav, ktery nastal pri stahovani</param>
        /// <param name="statusText">textova verze stavu</param>
        /// <param name="suggestedWaiting">doporuceny cas, ktery by se melo cekat, nez provede dalsi pokus o stahnuti souboru</param>
        public TimeoutStatusChangeEventArgs(DownloadStatus status, string statusText, int suggestedWaiting)
            : base(status, statusText)
        {
            this.suggestedWaiting = suggestedWaiting;
        }

        /// <summary>
        /// Vraci doporuceny cas, ktery by se melo cekat, nez provede dalsi pokus o stahnuti souboru
        /// </summary>
        public int SuggestedWaiting
        {
            get { return suggestedWaiting; }
        }
    }

    /// <summary>
    /// Trida zastresujici hodnotu argumentu u udalosti DownloadStatusChange s urcenim casu pro wait
    /// </summary>
    public class WaitingStatusChangeEventArgs : TimeoutStatusChangeEventArgs
    {
        /// <summary>
        /// Konstruktor tridy
        /// </summary>
        /// <param name="status">stav, ktery nastal pri stahovani</param>
        /// <param name="statusText">textova verze stavu</param>
        /// <param name="suggestedWaiting">doporuceny cas, ktery by se melo cekat, nez provede dalsi krok v procesu stahovani</param>
        public WaitingStatusChangeEventArgs(DownloadStatus status, string statusText, int suggestedWaiting)
            : base(status, statusText, suggestedWaiting) {}
    }

    /// <summary>
    /// Trida zastresujici hodnotu argumentu u udalosti DownloadFailed
    /// </summary>
    public class DownloadFailedEventArgs : DownloadStatusChangedEventArgs
    {
        private readonly string reason = "";

        /// <summary>
        /// Konstruktor tridy
        /// </summary>
        /// <param name="status">stav, ktery nastal pri stahovani</param>
        /// <param name="statusText">textova verze stavu</param>
        /// <param name="reason">duvod, proc bylo neuspesne stahnuti</param>
        public DownloadFailedEventArgs(DownloadStatus status, string statusText, string reason)
            : base(status, statusText)
        {
            this.reason = reason;
        }

        /// <summary>
        /// Vraci duvod, proc bylo neuspesne stahnuti
        /// </summary>
        public string Reason
        {
            get { return reason; }
        }
    }

    /// <summary>
    /// Trida zastresujici hodnotu argumentu u udalosti PositionChanged
    /// </summary>
    public class PositionChangedEventArgs : EventArgs
    {
        private readonly long position;
        private readonly long size;

        /// <summary>
        /// Konstruktor tridy
        /// </summary>
        /// <param name="position">celkova velikost jiz stahnutych bajtu u souboru</param>
        /// <param name="size">velikost souboru</param>
        public PositionChangedEventArgs(long position, long size)
        {
            this.position = position;
            this.size = size;
        }

        /// <summary>
        /// Vraci celkovou velikost jiz stahnutych bajtu u souboru
        /// </summary>
        public long Position
        {
            get { return position; }
        }

        /// <summary>
        /// Vraci velikost souboru
        /// </summary>
        public long Size
        {
            get { return size; }
        }

        /// <summary>
        /// Vraci procentualni vyjadreni stavu stahovaneho souboru
        /// </summary>
        public int Percents
        {
            get { return (size == 0) ? 0 : (int) (((float) position/size)*100); }
        }
    }
}